#include <dfsch-ext/tcltk.h>
#include <dfsch/magic.h>
#include <dfsch/util.h>

typedef struct interpreter_t {
  dfsch_type_t* type;
  Tcl_Interp* interpreter;
  int active;
} interpreter_t;

dfsch_type_t dfsch_ext_tcl_interpreter_type = {
  DFSCH_STANDARD_TYPE,
  NULL,
  sizeof(interpreter_t),
  "tcltk:interpreter",
  NULL,
  NULL,
  NULL,
  NULL
};

typedef struct command_wrapper_t {
  dfsch_type_t* type;
  char* name;
  dfsch_object_t* interpreter;
} command_wrapper_t;

static char* command_wrapper_write(command_wrapper_t* cw, 
                                   int depth, int readable){
  return dfsch_saprintf("#<command-wrapper 0x%x %s>", cw, cw->name);
}

static dfsch_object_t* command_wrapper_apply(command_wrapper_t* cw,
                                             dfsch_object_t* args,
                                             dfsch_tail_escape_t* esc){
  Tcl_Interp* interpreter;

  if (cw->interpreter){
    interpreter = dfsch_ext_tcl_interpreter(cw->interpreter);
  } else {
    DFSCH_EXT_TCL_INTERPRETER_ARG(args, interpreter);
  }

  return dfsch_ext_tcl_eval(interpreter,
                            dfsch_stracat(cw->name,
                                          dfsch_ext_tcl_quote_list(args)));
}

dfsch_type_t dfsch_ext_tcl_command_wrapper_type = {
  DFSCH_STANDARD_TYPE,
  NULL,
  sizeof(command_wrapper_t),
  "tcltk:command-wrapper",
  NULL,
  (dfsch_type_write_t)command_wrapper_write,
  (dfsch_type_apply_t)command_wrapper_apply,
  NULL
};


Tcl_Interp* dfsch_ext_tcl_interpreter(dfsch_object_t* obj){
  if (DFSCH_TYPE_OF(obj) != DFSCH_EXT_TCL_INTERPRETER_TYPE){
    dfsch_error("tcltk:not-a-interpreter", obj);
  }
  if (!((interpreter_t*)obj)->active){
    dfsch_error("tcltk:interpreter-already-destroyed", obj);
  }

  return ((interpreter_t*)obj)->interpreter;
}

static void interpreter_finalizer(interpreter_t* i, void* cd){
  if (i->active){
    i->active = 0;
    Tcl_DeleteInterp(i->interpreter);
  }
}

dfsch_object_t* dfsch_ext_tcl_make_interpreter(Tcl_Interp* interp){
  interpreter_t* i 
    = (interpreter_t*)dfsch_make_object(DFSCH_EXT_TCL_INTERPRETER_TYPE);

  i->interpreter = interp;
  i->active = 1;
  GC_register_finalizer(i, (GC_finalization_proc)interpreter_finalizer,
                        NULL, NULL, NULL);

  return (dfsch_object_t*)i;
}

dfsch_object_t* dfsch_ext_tcl_create_interpreter(){
  Tcl_Interp* i = Tcl_CreateInterp();
  if (Tcl_Init(i) == TCL_ERROR){
    dfsch_ext_tcl_error(i);
  }
  if (Tk_Init(i) == TCL_ERROR){
    dfsch_ext_tcl_error(i);
  }
  return dfsch_ext_tcl_make_interpreter(i);
}
void dfsch_ext_tcl_destroy_interpreter(dfsch_object_t* obj){
  if (DFSCH_TYPE_OF(obj) != DFSCH_EXT_TCL_INTERPRETER_TYPE){
    dfsch_error("tcltk:not-a-interpreter", obj);
  }
  if (!((interpreter_t*)obj)->active){
    dfsch_error("tcltk:interpreter-already-destroyed", obj);
  }
  
  ((interpreter_t*)obj)->active = 0;
  Tcl_DeleteInterp(((interpreter_t*)obj)->interpreter);
}

dfsch_object_t* dfsch_ext_tcl_wrap_command(char* name,
                                           dfsch_object_t* interp){
  command_wrapper_t* cw 
    = (command_wrapper_t*)dfsch_make_object(DFSCH_EXT_TCL_COMMAND_WRAPPER_TYPE);

  if (interp){
    if (DFSCH_TYPE_OF(interp) != DFSCH_EXT_TCL_INTERPRETER_TYPE){
      dfsch_error("tcltk:not-a-interpreter", interp);
    }
    if (!((interpreter_t*)interp)->active){
      dfsch_error("tcltk:interpreter-already-destroyed", interp);
    }
  }

  cw->interpreter = (interpreter_t*)interp;
  cw->name = name;

  return (dfsch_object_t*)cw;
}

typedef struct {
  dfsch_object_t* proc;
} command_context_t;

static int command_proc(command_context_t* ctx,
                        Tcl_Interp* interp,
                        int argc, char** argv){
  dfsch_object_t *head; 
  dfsch_object_t *cur;
  dfsch_object_t *res; 
  int i;
  int ret;

  head = cur = dfsch_multicons(argc);

  for(i = 0; i < argc; ++i){
    DFSCH_FAST_CAR(cur) = dfsch_make_string_cstr(argv[i]);
    cur = DFSCH_FAST_CDR(cur);
  }

  DFSCH_TRY {
    res = dfsch_apply(ctx->proc, head);
    Tcl_SetResult(interp, dfsch_obj_write(res, 100, 0), TCL_VOLATILE);
    ret = TCL_OK;
  } DFSCH_CATCH(ex) {
    Tcl_SetResult(interp, dfsch_obj_write(ex, 100, 0), TCL_VOLATILE);
    ret = TCL_ERROR;
  } DFSCH_END_TRY;

  return ret;
}

void dfsch_ext_tcl_create_command(Tcl_Interp* interp, 
                                  char* name, 
                                  dfsch_object_t* proc){
  command_context_t* ctx = GC_NEW_UNCOLLECTABLE(command_context_t);

  ctx->proc = proc;

  Tcl_CreateCommand(interp, name, command_proc, ctx, GC_free);
}

void dfsch_ext_tcl_error(Tcl_Interp* interp){
  dfsch_error("tcltk:error", 
              dfsch_make_string_cstr(Tcl_GetStringResult(interp)));
}

char* dfsch_ext_tcl_eval(Tcl_Interp* interp, char* string){
  if (Tcl_Eval(interp, string) == TCL_ERROR){
    dfsch_ext_tcl_error(interp);
  }
  return dfsch_stracpy(Tcl_GetStringResult(interp));
}

char* dfsch_ext_tcl_quote(char* str){
  char* ret;
  char* out;
  ret = out = GC_MALLOC_ATOMIC(strlen(str)*2+1);
  while (*str){
    switch (*str){
    case '\\':
      (*out++) = '\\';
      (*out++) = '\\';
      str++;
      break;
    case ' ':
      (*out++) = '\\';
      (*out++) = ' ';
      str++;
      break;
    case '\t':
      (*out++) = '\\';
      (*out++) = 't';
      str++;
      break;
    case '\n':
      (*out++) = '\\';
      (*out++) = 'n';
      str++;
      break;
    case '$':
      (*out++) = '\\';
      (*out++) = '$';
      str++;
      break;
    default:
      (*out++) = (*str++);
    }
  }

  (*out) = '\0';
  return ret;
}

char* dfsch_ext_tcl_quote_list(dfsch_object_t* list){
  dfsch_str_list_t* sl = dfsch_sl_create();
  dfsch_object_t* i;

  while (DFSCH_PAIR_P(list)){
    dfsch_sl_append(sl, " ");
    i = DFSCH_FAST_CAR(list);
    if (dfsch_string_p(i)){
      dfsch_sl_append(sl, dfsch_ext_tcl_quote(dfsch_string_to_cstr(i)));
    } else {
      dfsch_sl_append(sl, dfsch_ext_tcl_quote(dfsch_obj_write(i, 10, 1)));      
    }
    list = DFSCH_FAST_CDR(list);
  }

  return dfsch_sl_value(sl);
}
